@@ -19,6 +19,9 @@ struct websocket_output {
 	GString *str;
 	size_t str_done;
 	enum lws_write_protocol protocol;
+	int http_status;
+	const char *content_type;
+	ssize_t content_length;
 };
 
 struct websocket_conn {
@@ -51,6 +54,8 @@ struct websocket_ng_buf {
 static GQueue websocket_vhost_configs;
 static struct lws_context *websocket_context;
 static GThreadPool *websocket_threads;
+static mutex_t websocket_callback_lock = MUTEX_STATIC_INIT;
+static mutex_t websocket_service_lock = MUTEX_STATIC_INIT;
 
 
 static struct websocket_message *websocket_message_new(struct websocket_conn *wc) {
@@ -103,27 +108,27 @@ static void __websocket_queue_raw(struct websocket_conn *wc, const char *msg, si
 
 // appends to output buffer without triggering a response
 void websocket_queue_raw(struct websocket_conn *wc, const char *msg, size_t len) {
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
 	__websocket_queue_raw(wc, msg, len);
-	mutex_unlock(&wc->lock);
 }
 
 
 // num bytes in output buffer
 size_t websocket_queue_len(struct websocket_conn *wc) {
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
+
 	size_t ret = 0;
 	for (GList *l = wc->output_q.head; l; l = l->next) {
 		struct websocket_output *wo = l->data;
 		ret += (wo->str->len - LWS_PRE);
 	}
-	mutex_unlock(&wc->lock);
+
 	return ret;
 }
 
 
 // adds data to output buffer (can be null) and optionally triggers specified response
-int websocket_write_raw(struct websocket_conn *wc, const char *msg, size_t len,
+void websocket_write_raw(struct websocket_conn *wc, const char *msg, size_t len,
 		enum lws_write_protocol protocol, bool done)
 {
 	mutex_lock(&wc->lock);
@@ -132,36 +137,63 @@ int websocket_write_raw(struct websocket_conn *wc, const char *msg, size_t len,
 	wo->protocol = protocol;
 	g_queue_push_tail(&wc->output_q, websocket_output_new());
 
-	if (done) {
-		lws_callback_on_writable(wc->wsi);
-		lws_cancel_service(websocket_context);
-	}
-
 	mutex_unlock(&wc->lock);
 
-	return 0;
+	if (done) {
+		// Sadly lws_callback_on_writable() doesn't do any internal
+		// locking, therefore we must protect it against a concurrently
+		// running lws_service(), as well as against other threads
+		// invoking lws_callback_on_writable().
+		//
+		// Acquire the callback lock first, which is normally unlocked,
+		// then wake up the service thread and try to break out of
+		// lws_service().  The service thread holds the service lock
+		// while lws_service() is executing and releases it as soon as
+		// lws_service() is done. We therefore try to acquire the
+		// service lock here, which blocks us until lws_service() is
+		// actually done. At this point the service thread will try to
+		// acquire the callback lock, which is still held by us here,
+		// and so the service thread will block until we are done
+		// calling lws_callback_on_writable(). Finally we release both
+		// locks, which allows the service thread to resume
+		// lws_service().
+		//
+		// The suggested approach of using
+		// LWS_CALLBACK_EVENT_WAIT_CANCELLED together with a queue and
+		// then calling lws_callback_on_writable() from the service
+		// thread is not usable as libwebsockets 2.0 doesn't support
+		// LWS_CALLBACK_EVENT_WAIT_CANCELLED.
+
+		mutex_lock(&websocket_callback_lock);
+		lws_cancel_service(websocket_context);
+
+		mutex_lock(&websocket_service_lock);
+		lws_callback_on_writable(wc->wsi);
+
+		mutex_unlock(&websocket_service_lock);
+		mutex_unlock(&websocket_callback_lock);
+	}
 }
 
 
 // adds data to output buffer (can be null) and triggers specified response: http or binary websocket
-int websocket_write_http_len(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
-	return websocket_write_raw(wc, msg, len, LWS_WRITE_HTTP, done);
+void websocket_write_http_len(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
+	websocket_write_raw(wc, msg, len, LWS_WRITE_HTTP, done);
 }
-int websocket_write_http(struct websocket_conn *wc, const char *msg, bool done) {
-	return websocket_write_http_len(wc, msg, msg ? strlen(msg) : 0, done);
+void websocket_write_http(struct websocket_conn *wc, const char *msg, bool done) {
+	websocket_write_http_len(wc, msg, msg ? strlen(msg) : 0, done);
 }
-int websocket_write_text(struct websocket_conn *wc, const char *msg, bool done) {
-	return websocket_write_raw(wc, msg, strlen(msg), LWS_WRITE_TEXT, done);
+void websocket_write_text(struct websocket_conn *wc, const char *msg, bool done) {
+	websocket_write_raw(wc, msg, strlen(msg), LWS_WRITE_TEXT, done);
 }
-int websocket_write_binary(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
-	return websocket_write_raw(wc, msg, len, LWS_WRITE_BINARY, done);
+void websocket_write_binary(struct websocket_conn *wc, const char *msg, size_t len, bool done) {
+	websocket_write_raw(wc, msg, len, LWS_WRITE_BINARY, done);
 }
 
 
 void websocket_write_next(struct websocket_conn *wc) {
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
 	g_queue_push_tail(&wc->output_q, websocket_output_new());
-	mutex_unlock(&wc->lock);
 }
 
 
@@ -176,7 +208,7 @@ static const char *websocket_echo_process(struct websocket_message *wm) {
 static void websocket_message_push(struct websocket_conn *wc, websocket_message_func_t func) {
 	ilogs(http, LOG_DEBUG, "Adding HTTP/WS message to processing queue");
 
-	mutex_lock(&wc->lock);
+	LOCK(&wc->lock);
 
 	struct websocket_message *wm = wc->wm;
 	assert(wm != NULL);
@@ -187,8 +219,6 @@ static void websocket_message_push(struct websocket_conn *wc, websocket_message_
 	g_thread_pool_push(websocket_threads, wc, NULL);
 
 	wc->wm = websocket_message_new(wc);
-
-	mutex_unlock(&wc->lock);
 }
 
 
@@ -219,55 +249,8 @@ static void websocket_process(void *p, void *up) {
 }
 
 
-static int websocket_dequeue(struct websocket_conn *wc) {
-	if (!wc)
-		return 0;
-
-	int is_http = 0;
-
-	mutex_lock(&wc->lock);
-	struct websocket_output *wo;
-	struct lws *wsi = wc->wsi;
-	while ((wo = g_queue_pop_head(&wc->output_q))) {
-		// used buffer slot?
-		if (wo->str) {
-			// allocate post-buffer
-			g_string_set_size(wo->str, wo->str->len + LWS_SEND_BUFFER_POST_PADDING);
-			size_t to_send = wo->str->len - wo->str_done - LWS_SEND_BUFFER_POST_PADDING;
-			if (to_send) {
-				if (to_send > 2000)
-					ilogs(http, LOG_DEBUG, "Writing %lu bytes to LWS", (unsigned long) to_send);
-				else
-					ilogs(http, LOG_DEBUG, "Writing back to LWS: '%.*s'",
-							(int) to_send, wo->str->str + wo->str_done);
-				size_t ret = lws_write(wsi, (unsigned char *) wo->str->str + wo->str_done,
-						to_send, wo->protocol);
-				if (ret != to_send)
-					ilogs(http, LOG_ERR, "Invalid LWS write: %lu != %lu",
-							(unsigned long) ret,
-							(unsigned long) to_send);
-				wo->str_done += ret;
-
-				if (wo->protocol == LWS_WRITE_HTTP)
-					is_http = 1;
-			}
-		}
-		websocket_output_free(wo);
-	}
-	g_queue_push_tail(&wc->output_q, websocket_output_new());
-
-	mutex_unlock(&wc->lock);
-
-	int ret = 0;
-	if (is_http)
-		if (lws_http_transaction_completed(wsi) == 1) // may destroy `wc`
-			ret = -1;
-
-	return ret;
-}
-
-static const char *websocket_do_http_response(struct websocket_conn *wc, int status, const char *content_type,
-		ssize_t content_length)
+static const char *__websocket_write_http_response(struct websocket_conn *wc, int status,
+		const char *content_type, ssize_t content_length)
 {
 	uint8_t buf[LWS_PRE + 2048], *start = &buf[LWS_PRE], *p = start,
 		*end = &buf[sizeof(buf) - LWS_PRE - 1];
@@ -291,29 +274,88 @@ static const char *websocket_do_http_response(struct websocket_conn *wc, int sta
 
 	return NULL;
 }
-int websocket_http_response(struct websocket_conn *wc, int status, const char *content_type,
+static int websocket_dequeue(struct websocket_conn *wc) {
+	if (!wc)
+		return 0;
+
+	int is_http = 0;
+
+	mutex_lock(&wc->lock);
+	struct websocket_output *wo;
+	struct lws *wsi = wc->wsi;
+	while ((wo = g_queue_pop_head(&wc->output_q))) {
+		// used buffer slot?
+		if (!wo->str)
+			goto next;
+
+		if (wo->http_status) {
+			const char *err = __websocket_write_http_response(wc, wo->http_status,
+					wo->content_type, wo->content_length);
+			if (err) {
+				ilogs(http, LOG_ERR, "Failed to write HTTP response headers: %s", err);
+				goto next;
+			}
+		}
+
+		// allocate post-buffer
+		g_string_set_size(wo->str, wo->str->len + LWS_SEND_BUFFER_POST_PADDING);
+		size_t to_send = wo->str->len - wo->str_done - LWS_SEND_BUFFER_POST_PADDING;
+		if (to_send) {
+			if (to_send > 2000)
+				ilogs(http, LOG_DEBUG, "Writing %lu bytes to LWS", (unsigned long) to_send);
+			else
+				ilogs(http, LOG_DEBUG, "Writing back to LWS: '%.*s'",
+						(int) to_send, wo->str->str + wo->str_done);
+			size_t ret = lws_write(wsi, (unsigned char *) wo->str->str + wo->str_done,
+					to_send, wo->protocol);
+			if (ret != to_send)
+				ilogs(http, LOG_ERR, "Invalid LWS write: %lu != %lu",
+						(unsigned long) ret,
+						(unsigned long) to_send);
+			wo->str_done += ret;
+
+			if (wo->protocol == LWS_WRITE_HTTP)
+				is_http = 1;
+		}
+
+next:
+		websocket_output_free(wo);
+	}
+	g_queue_push_tail(&wc->output_q, websocket_output_new());
+
+	mutex_unlock(&wc->lock);
+
+	int ret = 0;
+	if (is_http)
+		if (lws_http_transaction_completed(wsi) == 1) // may destroy `wc`
+			ret = -1;
+
+	return ret;
+}
+
+void websocket_http_response(struct websocket_conn *wc, int status, const char *content_type,
 		ssize_t content_length)
 {
-	const char *err = websocket_do_http_response(wc, status, content_type, content_length);
-	if (!err)
-		return 0;
-	ilogs(http, LOG_ERR, "Failed to write HTTP response headers: %s", err);
-	return -1;
+	LOCK(&wc->lock);
+
+	struct websocket_output *wo = g_queue_peek_tail(&wc->output_q);
+
+	wo->http_status = status;
+	wo->content_type = content_type;
+	wo->content_length = content_length;
 }
-const char *websocket_http_complete(struct websocket_conn *wc, int status, const char *content_type,
+void websocket_http_complete(struct websocket_conn *wc, int status, const char *content_type,
 		ssize_t content_length, const char *content)
 {
-	if (websocket_http_response(wc, status, content_type, content_length))
-		return "Failed to write response HTTP headers";
-	if (websocket_write_http(wc, content, true))
-		return "Failed to write pong response";
-	return NULL;
+	websocket_http_response(wc, status, content_type, content_length);
+	websocket_write_http(wc, content, true);
 }
 
 
 static const char *websocket_http_ping(struct websocket_message *wm) {
 	ilogs(http, LOG_DEBUG, "Respoding to GET /ping");
-	return websocket_http_complete(wm->wc, 200, "text/plain", 5, "pong\n");
+	websocket_http_complete(wm->wc, 200, "text/plain", 5, "pong\n");
+	return NULL;
 }
 
 
@@ -350,7 +392,8 @@ static const char *websocket_http_metrics(struct websocket_message *wm) {
 		g_string_append_printf(outp, " %s\n", m->value_short);
 	}
 
-	return websocket_http_complete(wm->wc, 200, "text/plain", outp->len, outp->str);
+	websocket_http_complete(wm->wc, 200, "text/plain", outp->len, outp->str);
+	return NULL;
 }
 
 
@@ -382,7 +425,8 @@ static const char *websocket_http_cli(struct websocket_message *wm) {
 
 	size_t len = websocket_queue_len(wm->wc);
 
-	return websocket_http_complete(wm->wc, 200, "text/plain", len, NULL);
+	websocket_http_complete(wm->wc, 200, "text/plain", len, NULL);
+	return NULL;
 }
 
 
@@ -412,8 +456,7 @@ static void websocket_ng_send_ws(str *cookie, str *body, const endpoint_t *sin,
 }
 static void websocket_ng_send_http(str *cookie, str *body, const endpoint_t *sin, void *p1) {
 	struct websocket_conn *wc = p1;
-	if (websocket_http_response(wc, 200, "application/x-rtpengine-ng", cookie->len + 1 + body->len))
-		ilogs(http, LOG_WARN, "Failed to write HTTP headers");
+	websocket_http_response(wc, 200, "application/x-rtpengine-ng", cookie->len + 1 + body->len);
 	websocket_queue_raw(wc, cookie->s, cookie->len);
 	websocket_queue_raw(wc, " ", 1);
 	websocket_queue_raw(wc, body->s, body->len);
@@ -468,6 +511,14 @@ static const char *websocket_http_ng(struct websocket_message *wm) {
 
 
 
+static const char *websocket_http_404(struct websocket_message *wm) {
+	ilogs(http, LOG_WARN, "Unhandled HTTP URI: '%s'", wm->uri);
+	websocket_http_complete(wm->wc, 404, "text/plain", 10, "not found\n");
+	return NULL;
+}
+
+
+
 static int websocket_http_get(struct websocket_conn *wc) {
 	struct websocket_message *wm = wc->wm;
 	const char *uri = wm->uri;
@@ -482,12 +533,8 @@ static int websocket_http_get(struct websocket_conn *wc) {
 		handler = websocket_http_cli;
 	else if (!strcmp(uri, "/metrics"))
 		handler = websocket_http_metrics;
-
-	if (!handler) {
-		ilogs(http, LOG_WARN, "Unhandled HTTP GET URI: '%s'", uri);
-		websocket_http_complete(wm->wc, 404, "text/plain", 10, "not found\n");
-		return 0;
-	}
+	else
+		handler = websocket_http_404;
 
 	websocket_message_push(wc, handler);
 	return 0;
@@ -551,12 +598,8 @@ static int websocket_http_body(struct websocket_conn *wc, const char *body, size
 		handler = websocket_http_ng;
 	else if (!strcmp(uri, "/admin") && wm->method == M_POST && wm->content_type == CT_JSON)
 		handler = websocket_janus_process;
-
-	if (!handler) {
-		ilogs(http, LOG_WARN, "Unhandled HTTP POST URI: '%s'", wm->uri);
-		websocket_http_complete(wm->wc, 404, "text/plain", 10, "not found\n");
-		return 0;
-	}
+	else
+		handler = websocket_http_404;
 
 	websocket_message_push(wc, handler);
 	return 0;
@@ -1066,8 +1109,16 @@ err:
 
 static void websocket_loop(void *p) {
 	ilogs(http, LOG_INFO, "Websocket listener thread running");
-	while (!rtpe_shutdown)
+	while (!rtpe_shutdown) {
+		// see websocket_write_raw() for locking logic
+
+		mutex_lock(&websocket_service_lock);
 		lws_service(websocket_context, 100);
+		mutex_unlock(&websocket_service_lock);
+
+		mutex_lock(&websocket_callback_lock);
+		mutex_unlock(&websocket_callback_lock);
+	}
 
 	websocket_cleanup();
 }
